#include <cstdio>
#include <cmath>

using namespace std;

#include "defs.hh"
#include "reader.hh"
#include "freq2.hh"
#include "crc32.h"

static double errortotal=0;
static unsigned errorcount=0;
static void UpdateErrorCounter(double error)
{
    errortotal += error;
    errorcount++;
}
static void ErrorSummary(void)
{
    double error = errortotal / errorcount;
    double suggestion = pow(2, error/12.0);
    fprintf(stderr,
        "Mean error: %g - suggest you add \"0 %g\" to your configuration file\n",
            error, 1.0/suggestion);
}

static class Wisdom
{
	typedef std::map<std::string, double> values_t;
	typedef std::map<unsigned, values_t> wisdom_t;
	wisdom_t wisdom;
	void Load()
	{
		FILE *fp = fopen(".spc-wisdom", "rt");
		if(!fp)return;
		char Buf[512];
		while(fgets(Buf,sizeof Buf,fp))
		{
			unsigned val=0;
			sscanf(Buf, "%X:", &val);
			const char *s = strchr(Buf,':');
			if(!s)continue; ++s;
			for(;;)
			{
				string key;
				while(*s != '=') key += *s++;
				++s;
				if(*s=='\n' || *s=='\r' || !*s)break;
				sscanf(s, "%lf", &wisdom[val][key]);
				while(*s && *s!=';') ++s;
				if(*s==';') ++s;
				if(*s=='\n' || *s=='\r' || !*s)break;
			}
		}
		fclose(fp);
	}
	void Save()
	{
		FILE *fp = fopen(".spc-wisdom", "wt");
		for(wisdom_t::const_iterator i=wisdom.begin(); i!=wisdom.end(); ++i)
		{
			fprintf(fp, "%X:", i->first);
			for(values_t::const_iterator j = i->second.begin(); j!=i->second.end(); ++j)
				fprintf(fp, "%s=%g;", j->first.c_str(), j->second);
			fprintf(fp, "\n");
		}
		fclose(fp);
	}
public:
	Wisdom() { Load(); }
	~Wisdom() { Save(); }
	
	bool haswisdom(unsigned id, const string &key) const
	{
	    wisdom_t::const_iterator j = wisdom.find(id);
	    if(j == wisdom.end()) return false;
	    values_t::const_iterator i = j->second.find(key);
	    if(i == j->second.end()) return false;
	    return true;
	}
	double getwisdom(unsigned id, const string &key) const
	{
	    wisdom_t::const_iterator j = wisdom.find(id);
	    if(j == wisdom.end()) return 0;
	    values_t::const_iterator i = j->second.find(key);
	    if(i == j->second.end()) return 0;
	    return i->second;
	}
	void setwisdom(unsigned id, const string &key, double val)
	{
		wisdom[id][key] = val;
	}
} Wisdom;

struct InstrData
{    
    int gmnum;
    double basefreq;
    bool initialized;
    
    InstrData(): gmnum(0),basefreq(1),initialized(false)
    {
    }
    
    void setup(int gm, double bfreq=1)
    {
        gmnum = gm;
        basefreq = bfreq;
        initialized = true;
    }
    
    unsigned getwisdomkey(unsigned insno) const
    {
    	unsigned smpid = Instrs[insno].sampleid;
    	const string &data = Samples[smpid].data;
    	crc32_t c = crc32_startvalue;
    	for(unsigned a=0; a<data.size(); ++a) c = crc32_update(c, data[a]);
    	return c;
    }
    void loadwisdom(unsigned insno)
    {
    	unsigned key = getwisdomkey(insno);
    	if(Wisdom.haswisdom(key, "gmnum"))
    	{
    		gmnum = Wisdom.getwisdom(key, "gmnum");
    		initialized = true;
    	}
    	if(Wisdom.haswisdom(key, "basefreq"))
    	{
    		basefreq = Wisdom.getwisdom(key, "basefreq");
    		initialized = true;
    	}
    	if(initialized)
    		fprintf(stderr, "Instrument %u: Loaded data: G%c%d, basefreq=%g (wavelen=%g)\n",
    			insno,
    			IsGM(gmnum) ? 'M' : 'P',
    			IsGM(gmnum) ? gmnum+1 : gmnum-128,
    			basefreq,
    			44100*2/basefreq);
    }
    void savewisdom(unsigned insno) const
    {
    	unsigned key = getwisdomkey(insno);
    	Wisdom.setwisdom(key, "gmnum",    gmnum);
    	Wisdom.setwisdom(key, "basefreq", basefreq);
    }
    
    void setupGM(int gm, double bfreq=1) { setup(GM(gm), bfreq); }
    void setupGP(int gp) { setup(GP(gp)); }
    
    inline double HzToNote(double hz) const
    {
        // This converts Hz to MIDI pitch spec.
        // Examples:
        //   440  Hz -> 69.0
        //   16.4 Hz -> 12.0
        //  12.5 kHz -> 127.0
        // Magical constants:
        //  17.31234049066756 = 12/log(2)
        //  36.37631656229592 = log(440)*12/log(2) - 69
        return 17.31234049066756 * log(hz) - 36.37631656229592;
    }
    double getnote(unsigned long freq) const
    {
        double n = freq / 65536.0;
        // n: 1.0 means 44100 samples per second
        //    2.0 means 88100 samples per second.
        
        //double h = 250;     // Assume this is the wave length
        //double 44100.0 / h; // So this is the frequency it will be played if n=1.
        //   ^ basefreq is this value.
        
        return HzToNote(n * basefreq);
    }
    int guessnote(unsigned long freq) const
    {
        int pitch = (int)(getnote(freq));
        if(pitch < 0) pitch = 0;
        if(pitch > 127) pitch = 127;
        
        return pitch;
    }
    int guessbend(unsigned long freq, int note) const
    {
        //return 0x2000;
        
        // 0x1000 = yksi puolinuotti
        double thisnote = getnote(freq);
        double error = thisnote - note;
        
        UpdateErrorCounter(error);
        
        int e = (int)(0x1000 * error);
        
        if(e < -0x2000) e = -0x2000;
        if(e >= 0x2000) e = 0x1FFF;
        
        return e + 0x2000;
    }
};
static map<unsigned, InstrData> InstruData;

typedef unsigned char BYTE;

static unsigned deltaticks = 1000, tempo = 500000;


/*   Tempo      = microseconds / quarternote
     Deltaticks = ticks / quarternote
     
     deltaticks * quarternote = ticks
     tempo * quarternote  = microseconds
     (quarternote =) ticks/deltaticks = microseconds/tempo
     
      ticks = microseconds*deltaticks/tempo
      microseconds = ticks*tempo/deltaticks

 */
struct MIDItrack
{
    vector<BYTE> data;
    
    BYTE running;
    double lasttimer, timer;
    
    MIDItrack() : running(0), lasttimer(0), timer(0)
    {
    }

    void add(BYTE c) { data.push_back(c); }
    void addvarlen(unsigned t)
    {
        if(t>>21)add(0x80 | ((t>>21) &127));
        if(t>>14)add(0x80 | ((t>>14) &127));
        if(t>>7) add(0x80 | ((t>>7)  &127));
        add(t&127);
    }
    void flush()
    {
        double scale = 500000;
        double timerdiff = (timer - lasttimer) * scale;
        unsigned ticks = (unsigned)(timerdiff * deltaticks / tempo);
        double elapsed = ((double)ticks * tempo / deltaticks) / scale;
        lasttimer += elapsed;
        
        addvarlen(ticks);
    }
    void addtext(BYTE n, const char *t)
    {
        flush();
        add(0xFF);
        add(n);   
        add(strlen(t));
        while(*t)add(*t++);
    }
    void markloopstart(void)
    {
        addtext(6, "loopStart");
    }
    void markloopend(void)
    {
        addtext(6, "loopEnd");
    }
    void midibyte(BYTE c)
    {
        if(c & 0x80) { flush(); }
        if(c & 0x80) { if(c == running)return; running = c; }
        add(c);
    }
};

static void WriteMIDI()
{
    vector<MIDItrack> tracks(8);
    vector<bool> trackempty(8, true);
    
    for(unsigned a=0; a<tracks.size(); ++a)
    {
#if 0
        tracks[a].flush();
        tracks[a].add(0xFF);
        tracks[a].add(0x58);
        tracks[a].add(0x04);
        tracks[a].add(4);   
        tracks[a].add(2);   
        tracks[a].add(24);    /* ticks/metro  */
        tracks[a].add(8);     /* 32nd per 1/4 */
#endif
        tracks[a].flush();
        tracks[a].add(0xFF);
        tracks[a].add(0x51);
        tracks[a].add(0x03);
        tracks[a].add((BYTE)(tempo >> 16));
        tracks[a].add((BYTE)(tempo >> 8)); 
        tracks[a].add((BYTE)(tempo));

        char Buf[64]; sprintf(Buf, "SNES track %u", a);
        tracks[a].addtext(4, Buf);
    }
    
    // Tnne nuotit
    
    bool ekatrack = false;
    
    /* Process each track */
    for(unsigned a=0; a<8; ++a)
    {
/*
        if(!
            (a==0)
          ) continue;
*/
        bool keyon = false;
        int ch=a;
        int note=0;
        int patch=0;
        int wheel=0x2000;
        unsigned curins=0;
        unsigned long freq=0;
        int lvol=0,rvol=0,volavg=0,lastpan=-1;
        int lastlvol=0,lastrvol=0,lastavg=0;
        int lastch=10;
        
        tracks[a].midibyte(0xB9);
           tracks[a].midibyte(7);
           tracks[a].midibyte(70); // set percussion volume to something nice

        const vector<Event> &events = Channels[a].events;
        for(unsigned e=0; e<events.size(); ++e)
        {
            const Event &ev = events[e];

            fprintf(stderr, "[%u]: time(%10g)  ev(%c)  vol(%4d %4d)  freq(%lu) ins(%u)",
                a, ev.time, ev.evtype, ev.l, ev.r, ev.freq, ev.ins);

            if(freq && ev.ins)
            {
                int note = InstruData[ev.ins].guessnote(freq);
                int oct = note/12;
                note %= 12;
                static const char notes[12*3+1]=
                "C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
                fprintf(stderr, " (%s%d)", notes+note*3, oct);
            }
            
            fprintf(stderr, "\n");
            
            tracks[a].timer = ev.time;
            
            switch(ev.evtype)
            {
                case 'K': // keyoff (no data)
                {
                    if(keyon)
                    {
                        tracks[a].midibyte(0x80 + ch);
                        tracks[a].midibyte(note);
                        tracks[a].midibyte(127); // average fade
                        keyon = false;
                    }
                    else
                        fprintf(stderr, "Twice keyoff at channel %u, time %g\n", a, ev.time);
                    break;
                }
                case 'O': // keyon (freq, .ins)
                {
                    if(!ev.ins) break; // instrument unknown, thus ignore!
                    
                    if(keyon)
                    {
                        tracks[a].midibyte(0x80 + ch);
                        tracks[a].midibyte(note);
                        tracks[a].midibyte(127); // average fade
                        keyon = false;
                        fprintf(stderr, "Twice keyon at channel %u, time %g\n", a, ev.time);
                    }
                    
                    const InstrData &ins = InstruData[curins = ev.ins];
                    
                    int gm = ins.gmnum;
                    
                    if(IsGP(gm))
                    {
                        ch   = 9;
                        note = GetGP(gm);
                    }
                    else
                    {
                        ch   = a;
                        note = ins.guessnote(freq);
                        gm   = GetGM(gm);
                        
                        if(patch != gm)
                        {
                            tracks[a].midibyte(0xC0+ch);
                            tracks[a].midibyte(patch = gm);
                        }
                    }

                    keyon = true;
                    
                    trackempty[a] = false;
                    if(ekatrack)
                    {
                        ekatrack = false;
                        tracks[a].addtext(2, "Generated using a SPC (snes music) to MIDI converter");
                        tracks[a].addtext(2, "Written by and copyright (C) 1992,2003 Bisqwit");
                        tracks[a].addtext(2, "http://iki.fi/bisqwit/");
                    }
                    
                    int pressure = 90;
                    if(ch == 9)pressure = volavg;
                    lastavg = volavg;

                    tracks[a].midibyte(0x90 + ch);
                    tracks[a].midibyte(note);
                    tracks[a].midibyte(pressure);
                    
                    if(ch != lastch)
                    {
                        lastch = ch;
                        lastlvol=lastrvol=999;
                    }

                    if(ch != 9) { goto Bend; }

                    break;
                }
                case 'F': // freq (.freq)
                {
                    freq = ev.freq;
                    if(keyon && ch != 9) // bending!
                    {
            Bend:       const InstrData &ins = InstruData[curins];
                        
                        int newwheel = ins.guessbend(freq, note);
                        
                        if(newwheel < 0) newwheel = 0;
                        if(newwheel > 0x3FFF) newwheel = 0x3FFF;
                        
                        if(newwheel != wheel)
                        {
                            wheel = newwheel;

                            tracks[a].midibyte(0xE0 + ch);
                            tracks[a].midibyte(wheel & 127);
                            tracks[a].midibyte(wheel >> 7);

                        }
                        goto SetVolume;
                    }
                    break;
                }
                case 'V': // pan, volumes  (.l, .r)
                {
                    lvol = ev.l * GLOBALVOLUME;
                    rvol = ev.r * GLOBALVOLUME;
                    volavg = (lvol+rvol)/2;
                    if(volavg<0)volavg=0;
                    if(volavg>127)volavg=127;
                    
SetVolume:          if(keyon && (lastlvol!=lvol || lastrvol!=rvol))
                    {
                        lastlvol = lvol;
                        lastrvol = rvol;
                        
                        if(ch == 9)
                        {
                            if(lastavg != volavg)
                            {
                                tracks[a].midibyte(0xA0 + ch);
                                tracks[a].midibyte(note);
                                tracks[a].midibyte(lastavg = volavg);
                            }
                        }
                        else
                        {
                            int max = lastlvol;
                            if(lastrvol > max) max = lastrvol;
                            
                            /* Scale the volume to midi level */
                            if(max < 0)max = 0; if(max > 127)max = 127;
                            
                            /* master volume */
                            tracks[a].midibyte(0xB0 + ch);
                            tracks[a].midibyte(7);
                            tracks[a].midibyte(max);
                            
                            /* pan position */
                            if(max > 0)
                            {
                                int panpos = 0x40 + (lastlvol-lastrvol) * 0x3F / (lastlvol + lastrvol);
                                
                                //fprintf(stderr, "Pan pos: %d (from %d,%d)\n", panpos, lastlvol, lastrvol);
                                
                                if(panpos != lastpan)
                                {
                                    tracks[a].midibyte(0xB0 + ch);
                                    tracks[a].midibyte(10);
                                    tracks[a].midibyte(lastpan = panpos);
                                }
                            }
                        }
                    }
                    break;
                }
            }
        }
    }
    
    // Trackit loppuivat.
    
    /* Erase unused tracks. */
    for(unsigned a=0,b=0; a<8; ++a)
        if(trackempty[a])
            tracks.erase(tracks.begin() + b);
        else
            ++b;
    
    /* Finish the tracks */
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        tracks[a].flush();
        tracks[a].add(0xFF);
        tracks[a].add(0x2F);
        tracks[a].add(0x00);
    }
    
    FILE *fp = fopen("midifile.mid", "wb");
    
    fwrite("MThd\0\0\0\6\0\1", 4+4+2, 1, fp);
    fputc(tracks.size() >> 8,  fp);
    fputc(tracks.size() & 255, fp);
    fputc(deltaticks >> 8,  fp);
    fputc(deltaticks & 255, fp);
    
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        const MIDItrack &tr = tracks[a];
    
        unsigned tracklen = tr.data.size();
        const BYTE *track = &tr.data[0];
        
        fwrite("MTrk", 1, 4, fp);
        fputc((BYTE)(tracklen>>24), fp);
        fputc((BYTE)(tracklen>>16), fp);
        fputc((BYTE)(tracklen>>8), fp); 
        fputc((BYTE)tracklen, fp);
    
        fwrite(track, 1, tracklen, fp);
    }
    
    fclose(fp);
}

static void LoadConf()
{
    FILE *fp = fopen("spctomidi.conf", "rt");
    char Buf[512];
    bool ok=true;
    
    if(!fp)
    {
    	perror("spctomidi.conf");
    	return;
    }
    
    double fixerror = 1.0;
    
    for(map<unsigned,Instr>::const_iterator i=Instrs.begin(); i!=Instrs.end(); ++i)
    {
        InstruData[i->first].loadwisdom(i->first);
    }
    
    while((fgets(Buf,sizeof Buf,fp)))
    {
        char *s;
        if((s = strchr(Buf, '#'))) { for(*s=0; --s >= Buf; *s=0)if(*s!=' ')break; }
        if(!Buf[0])continue;
        if(Buf[0]=='-') { ok=false; continue; }
        if(Buf[0]=='+') { ok=true; continue; }
        if(!ok)continue;
        if(Buf[0] == '0')
        {
            double tmp=1.0;
            sscanf(Buf, "0 %lf", &tmp);
            fixerror *= tmp;
            continue;
        }
        if(Buf[0] < '0' || Buf[0] > '9')continue;
        unsigned insno=0;
        char gpnum='?';
        int gm=0;
        sscanf(Buf, "%u G%c%d %s", &insno, &gpnum, &gm, Buf);
        if(gpnum=='M')
        {
            double kerroin=1;
            s = Buf; while(*s == ' ')++s;
        	unsigned rate = 44100;
            if(*s=='g')
            {
            	fprintf(stderr, "insno=%u, sampleid=%u\n",
            		insno, Instrs[insno].sampleid);
                double wavelen = GetWaveLen(Instrs[insno].sampleid, rate);
                kerroin = rate/wavelen * 2.0;
            }
            else
            {
                sscanf(s, "%lf", &kerroin);
                kerroin = rate/kerroin * 2.0;
            }
            InstruData[insno].setupGM(gm, kerroin * fixerror);
        }
        else if(gpnum=='P')
        {
            InstruData[insno].setupGP(gm);
        }
        InstruData[insno].savewisdom(insno);
    }
    for(map<unsigned,Instr>::const_iterator i=Instrs.begin(); i!=Instrs.end(); ++i)
    	if(!InstruData[i->first].initialized)
    		fprintf(stderr, "Warning: Instrument %u uninitialized!\n", i->first);
    
    fclose(fp);
}

int main(void)
{
    ReadLog(stdin);
    
    LoadConf();

    WriteMIDI();
    
    ErrorSummary();
    
    return 0;
}
